type bool =
    True
  | False
;;

type nat =
    O
  | S of nat
;;

type ('a, 'b) prod =
    Pair of 'a * 'b
;;

let prod_rec f = function
  Pair (x0, x) -> f x0 x
;;

type sumbool =
    Left_ren
  | Right_ren
;;

let sumbool_rec f f0 = function
  Left_ren -> f
| Right_ren -> f0
;;

type 'a sumor =
    Inleft of 'a
  | Inright
;;

let rec plus n m =
  match n with
    O -> m
  | S p -> S (plus p m)
;;

let acc_rec f =
  let rec acc_rec x =
    f x (fun y -> acc_rec y)
  in acc_rec
;;

let well_founded_induction h0 a0 =
  acc_rec (fun x h2 -> h0 x h2) a0
;;

let rec minus n m =
  match n with
    O -> O
  | S k -> (match m with
              O -> S k
            | S l -> minus k l)
;;

let eq_nat_dec n =
  let rec f1 = function
    O -> (fun m ->
      let rec f3 = function
        O -> Left_ren
      | S n1 -> Right_ren
      in f3 m)
  | S n0 -> (fun m ->
      let rec f2 = function
        O -> Right_ren
      | S n3 -> sumbool_rec Left_ren Right_ren (f1 n0 n3)
      in f2 m)
  in f1 n
;;

let lt_eq_lt_dec n =
  let rec f1 = function
    O -> (fun m ->
      let rec f3 = function
        O -> Inleft Right_ren
      | S n1 -> Inleft Left_ren
      in f3 m)
  | S n0 -> (fun m ->
      let rec f2 = function
        O -> Inright
      | S n3 ->
          (match f1 n0 n3 with
             Inleft x ->
               sumbool_rec (Inleft Left_ren) (Inleft Right_ren) x
           | Inright -> Inright)
      in f2 m)
  in f1 n
;;

let le_lt_dec n =
  let rec f1 = function
    O -> (fun m -> Left_ren)
  | S n0 -> (fun m ->
      let rec f2 = function
        O -> Right_ren
      | S n3 -> sumbool_rec Left_ren Right_ren (f1 n0 n3)
      in f2 m)
  in f1 n
;;

let le_lt_eq_dec n m =
  match lt_eq_lt_dec n m with
    Inleft x -> x
  | Inright -> failwith "False_rec"
;;

let max h' =
  let rec f1 = function
    O -> (fun n1 -> n1)
  | S n0 -> (fun h0 -> match h0 with
                         O -> S n0
                       | S n2 -> S (f1 n0 n2))
  in f1 h'
;;

type mon =
    N_0
  | C_n of nat * nat * mon
;;

let pmon1 h = function
  N_0 -> O
| C_n (i, n, j) -> n
;;

let pmon2 k = function
  N_0 -> N_0
| C_n (l, m0, m') -> m'
;;

let gen_mon d =
  let rec f1 = function
    O -> (fun n0 -> N_0)
  | S n0 -> (fun n' ->
      match n' with
        O -> C_n (n0, (S O), (f1 n0 n0))
      | S n'' -> C_n (n0, O, (f1 n0 n'')))
  in f1 d
;;

let mult_mon d =
  let rec f1 = function
    O -> (fun h' h'0 -> N_0)
  | S n0 -> (fun s1 s2 ->
      C_n (n0, (plus (pmon1 (S n0) s1) (pmon1 (S n0) s2)),
           (f1 n0 (pmon2 (S n0) s1) (pmon2 (S n0) s2))))
  in f1 d
;;

let zero_mon d =
  let rec f1 = function
    O -> N_0
  | S n0 -> C_n (n0, O, (f1 n0))
  in f1 d
;;

let div_mon d =
  let rec f1 = function
    O -> (fun h' h'0 -> N_0)
  | S n0 -> (fun s1 s2 ->
      C_n (n0, (minus (pmon1 (S n0) s1) (pmon1 (S n0) s2)),
           (f1 n0 (pmon2 (S n0) s1) (pmon2 (S n0) s2))))
  in f1 d
;;

let div_mon_clean d =
  let rec f1 = function
    O -> (fun h' h'0 -> Pair (N_0, True))
  | S n0 -> (fun s1 s2 ->
      match le_lt_dec (pmon1 (S n0) s2) (pmon1 (S n0) s1) with
        Left_ren ->
          (match f1 n0 (pmon2 (S n0) s1) (pmon2 (S n0) s2) with
             Pair (res, b) ->
               Pair ((C_n (n0,
                           (minus (pmon1 (S n0) s1) (pmon1 (S n0) s2)),
                           res)), b))
      | Right_ren -> Pair (s1, False))
  in f1 d
;;

let eqmon_dec d =
  let rec f1 = function
    O -> (fun x y -> Left_ren)
  | S n0 -> (fun x y ->
      sumbool_rec
        (sumbool_rec Left_ren Right_ren
          (f1 n0 (pmon2 (S n0) x) (pmon2 (S n0) y))) Right_ren
        (eq_nat_dec (pmon1 (S n0) x) (pmon1 (S n0) y)))
  in f1 d
;;

let ppcm_mon d =
  let rec f1 = function
    O -> (fun m1 m2 -> N_0)
  | S n0 -> (fun s1 s2 ->
      C_n (n0, (max (pmon1 (S n0) s1) (pmon1 (S n0) s2)),
           (f1 n0 (pmon2 (S n0) s1) (pmon2 (S n0) s2))))
  in f1 d
;;

let letP h h' =
  h' h
;;

let orderc_dec n a =
  let rec f1 n = function
    N_0 -> (fun b -> Inright)
  | C_n (d, n0, m0) -> (fun b ->
      match f1 d m0 (pmon2 (S d) b) with
        Inleft h'0 ->
          (match h'0 with
             Left_ren -> Inleft Left_ren
           | Right_ren -> Inleft Right_ren)
      | Inright ->
          (match lt_eq_lt_dec n0 (pmon1 (S d) b) with
             Inleft x -> sumbool_rec (Inleft Right_ren) Inright x
           | Inright -> Inleft Left_ren))
  in f1 n a
;;

let degc n h' =
  let rec f1 n = function
    N_0 -> O
  | C_n (d, n0, m0) -> plus n0 (f1 d m0)
  in f1 n h'
;;

let total_orderc_dec n a b =
  letP (degc n a) (fun u ->
    letP (degc n b) (fun u0 ->
      match le_lt_dec u u0 with
        Left_ren ->
          (match le_lt_eq_dec u u0 with
             Left_ren -> Inleft Left_ren
           | Right_ren ->
               (match orderc_dec n a b with
                  Inleft h'3 ->
                    (match h'3 with
                       Left_ren -> Inleft Left_ren
                     | Right_ren -> Inleft Right_ren)
                | Inright -> Inright))
      | Right_ren -> Inleft Right_ren))
;;

let f = function
  Pair (a1, d1) -> d1
;;

let zeroP_dec b0 eqD_dec = function
  Pair (b, h') -> eqD_dec b b0
;;

let plusA plusD = function
  Pair (b2, c2) -> (fun y ->
    match y with
      Pair (b3, c3) -> Pair ((plusD b2 b3), c2))
;;

let multA multD plusC = function
  Pair (b2, c2) -> (fun h1' ->
    match h1' with
      Pair (b3, c3) -> Pair ((multD b2 b3), (plusC c2 c3)))
;;

let invA invD = function
  Pair (b2, c2) -> Pair ((invD b2), c2)
;;

let minusA minusD = function
  Pair (b2, c2) -> (fun h' ->
    match h' with
      Pair (b3, c3) -> Pair ((minusD b2 b3), c2))
;;

let divA divD minusC = function
  Pair (b2, c2) -> (fun h' ->
    match h' with
      Pair (b3, c3) -> Pair ((divD b2 b3), (minusC c2 c3)))
;;

let divA_dec b0 multD plusC divD minusC minusC_clean a b =
  match a with
    Pair (x0, x) ->
      (match b with
         Pair (b2, c2) ->
           (match minusC_clean x c2 with
              Pair (c0, b4) ->
                (match b4 with
                   True -> Left_ren
                 | False -> Right_ren)))
;;

let ppc d1 maxC = function
  Pair (b2, c2) -> (fun h' ->
    match h' with
      Pair (b3, c3) -> Pair (d1, (maxC c2 c3)))
;;

type 'a list =
    Nil
  | Cons of 'a * ('a list)
;;

let list_rec f0 f1 =
  let rec f2 = function
    Nil -> f0
  | Cons (a0, l0) -> f1 a0 l0 (f2 l0)
  in f2
;;

let eqB_dec eqA_dec f0 x y =
  eqA_dec (f0 x) (f0 y)
;;

let ltB_dec ltA_dec f0 x y =
  ltA_dec (f0 x) (f0 y)
;;

let pluspf f0 zeroP_dec0 ltC_dec plusA0 l1 l2 =
  well_founded_induction (fun x ->
    match x with
      Pair (p, q) ->
        (match p with
           Nil -> (fun h' -> q)
         | Cons (a0, m1) ->
             (match q with
                Nil -> (fun h' -> Cons (a0, m1))
              | Cons (a2, m2) -> (fun h' ->
                  match ltB_dec ltC_dec f0 a0 a2 with
                    Inleft p0 ->
                      (match p0 with
                         Left_ren ->
                           Cons (a2, (h' (Pair ((Cons (a0, m1)), m2))))
                       | Right_ren ->
                           Cons (a0, (h' (Pair (m1, (Cons (a2, m2)))))))
                  | Inright ->
                      letP (plusA0 a0 a2) (fun letA ->
                        match zeroP_dec0 letA with
                          Left_ren -> h' (Pair (m1, m2))
                        | Right_ren ->
                            Cons (letA, (h' (Pair (m1, m2)))))))))
    (Pair (l1, l2))
;;

let splus f0 zeroP_dec0 ltC_dec plusA0 sp1 sp2 =
  pluspf f0 zeroP_dec0 ltC_dec plusA0 sp2 sp1
;;

let multm_lm multA0 a0 p =
  list_rec Nil (fun b p1 p'1 -> Cons ((multA0 a0 b), p'1)) p
;;

let tmultm_lm zeroP_dec0 multA0 a0 =
  match zeroP_dec0 a0 with
    Left_ren -> (fun h' -> Nil)
  | Right_ren -> (fun p -> multm_lm multA0 a0 p)
;;

let minuspf f0 zeroP_dec0 ltC_dec multA0 invA0 a0 minusA0 l1 l2 =
  well_founded_induction (fun x ->
    match x with
      Pair (l3, l0) ->
        (match l3 with
           Nil -> (fun h' -> multm_lm multA0 (invA0 a0) l0)
         | Cons (a2, m1) ->
             (match l0 with
                Nil -> (fun h' -> Cons (a2, m1))
              | Cons (a3, m2) -> (fun h' ->
                  match ltB_dec ltC_dec f0 a2 a3 with
                    Inleft p ->
                      (match p with
                         Left_ren ->
                           Cons ((invA0 a3),
                                 (h' (Pair ((Cons (a2, m1)), m2))))
                       | Right_ren ->
                           Cons (a2, (h' (Pair (m1, (Cons (a3, m2)))))))
                  | Inright ->
                      letP (minusA0 a2 a3) (fun u ->
                        match zeroP_dec0 u with
                          Left_ren -> h' (Pair (m1, m2))
                        | Right_ren -> Cons (u, (h' (Pair (m1, m2)))))))))
    (Pair (l1, l2))
;;

let spminusf f0 zeroP_dec0 ltC_dec multA0 invA0 a0 minusA0 divA0 a2 b p q =
  minuspf f0 zeroP_dec0 ltC_dec multA0 invA0 a0 minusA0 p
    (multm_lm multA0 (divA0 a2 b) q)
;;

let spolyf f0 zeroP_dec0 ltC_dec multA0 invA0 a0 minusA0 divA0 ppc0 = function
  Nil -> (fun p2 -> Nil)
| Cons (a2, p11) -> (fun p2 ->
    match p2 with
      Nil -> Nil
    | Cons (b, p22) ->
        letP (ppc0 a2 b) (fun u ->
          minuspf f0 zeroP_dec0 ltC_dec multA0 invA0 a0 minusA0
            (multm_lm multA0 (divA0 u a2) p11)
            (multm_lm multA0 (divA0 u b) p22)))
;;

let multpf f0 zeroP_dec0 ltC_dec plusA0 multA0 =
  let rec multpf p q =
    match p with
      Nil -> Nil
    | Cons (a0, p') ->
        pluspf f0 zeroP_dec0 ltC_dec plusA0 (multm_lm multA0 a0 q)
          (multpf p' q)
  in multpf
;;

let smult f0 zeroP_dec0 ltC_dec plusA0 multA0 sp1 sp2 =
  multpf f0 zeroP_dec0 ltC_dec plusA0 multA0 sp2 sp1
;;

type 'poly cpRes =
    Keep of ('poly list)
  | DontKeep of ('poly list)
;;

let addRes i = function
  Keep h'0 -> Keep (Cons (i, h'0))
| DontKeep h'0 -> DontKeep (Cons (i, h'0))
;;

let slice divP_dec0 ppc0 foreigner_dec i a q =
  list_rec
    (match foreigner_dec i a with
       Left_ren -> DontKeep Nil
     | Right_ren -> Keep Nil) (fun b q1 rec_ren ->
    match divP_dec0 (ppc0 i a) b with
      Left_ren -> DontKeep (Cons (b, q1))
    | Right_ren ->
        (match divP_dec0 (ppc0 i b) a with
           Left_ren -> rec_ren
         | Right_ren -> addRes b rec_ren)) q
;;

let genPcPf spoly divP_dec0 addS ppc0 foreigner_dec i aP q =
  well_founded_induction (fun aP0 ->
    match aP0 with
      Nil -> (fun h' r -> r)
    | Cons (a, l1) -> (fun rec_ren l ->
        match slice divP_dec0 ppc0 foreigner_dec i a l1 with
          Keep l2 -> addS (spoly i a) (rec_ren l2 l)
        | DontKeep l2 -> rec_ren l2 l)) aP q
;;

let genOCPf spoly divP_dec0 addS ppc0 foreigner_dec h' =
  list_rec Nil (fun a l rec_ren ->
    genPcPf spoly divP_dec0 addS ppc0 foreigner_dec a l rec_ren) h'
;;

let pbuchf spoly divP_dec0 addL addS zerop_dec ppc0 nf foreigner_dec pq =
  well_founded_induction (fun x ->
    prod_rec (fun p q ->
      match q with
        Nil -> (fun h' -> p)
      | Cons (a, q2) -> (fun rec_ren ->
          letP (nf a p) (fun a0 ->
            match zerop_dec a0 with
(*******)
              Left_ren -> (print_string "0";flush stdout;
                           rec_ren (Pair (p, q2)))
            | Right_ren ->(print_string "n";flush stdout;
(*******)
                rec_ren
                  (Pair ((addL a0 p),
                         (genPcPf spoly divP_dec0 addS ppc0
                           foreigner_dec a0 p q2))))))) x) pq
;;

let buch spoly divP_dec0 addL addS zerop_dec ppc0 nf foreigner_dec p =
  pbuchf spoly divP_dec0 addL addS zerop_dec ppc0 nf foreigner_dec
    (Pair (p, (genOCPf spoly divP_dec0 addS ppc0 foreigner_dec p)))
;;

let addEnd f0 zeroP_dec0 a0 h'0 =
  list_rec (Cons (a0, Nil)) (fun b l1 rec_ren -> Cons (b, rec_ren)) h'0
;;

let spolyp f0 zeroP_dec0 ltC_dec multA0 invA0 a0 minusA0 divA0 ppc0 p q =
  spolyf f0 zeroP_dec0 ltC_dec multA0 invA0 a0 minusA0 divA0 ppc0 q p
;;

let zerop_dec f0 zeroP_dec0 = function
  Nil -> Left_ren
| Cons (a0, l) -> Right_ren
;;

let divp_dec f0 zeroP_dec0 multA0 divA0 divA_dec0 a0 = function
  Nil -> (match a0 with
            Nil -> Right_ren
          | Cons (a2, l) -> Right_ren)
| Cons (x0, x) ->
    (match a0 with
       Nil -> Right_ren
     | Cons (a2, l) -> sumbool_rec Left_ren Right_ren (divA_dec0 a2 x0))
;;

let ppcp f0 zeroP_dec0 ppc0 = function
  Nil -> (fun h'1 -> Nil)
| Cons (a0, l) -> (fun h'1 ->
    match h'1 with
      Nil -> Nil
    | Cons (a2, l0) -> Cons ((ppc0 a0 a2), Nil))
;;

let nf f0 zeroP_dec0 ltC_dec multA0 invA0 a0 minusA0 divA0 reduce1a unit p l =
  letP (reduce1a l p) (fun u -> multm_lm multA0 (unit u) u)
;;

let foreigner_dec f0 zeroP_dec0 eqCoef_dec multA0 ppc0 a0 = function
  Nil -> Left_ren
| Cons (a2, l) ->
    (match a0 with
       Nil -> Left_ren
     | Cons (a3, l0) ->
         eqB_dec eqCoef_dec f0 (ppc0 a2 a3) (multA0 a2 a3))
;;

let rec  append l l2 = match l with
 Nil -> l2
| (Cons (a,p)) -> (Cons (a, (append p l2)));;

let rec redd l l2 red = match l with 
|  Nil -> Nil
| (Cons (a,p)) -> let l22= append l2 p in
                  let pol = red a l22 in
                 ( match pol with 
                  Nil -> redd p l2 red
                 | _ -> (Cons (pol, redd p (Cons (pol,l2)) red)))
;;

let buchf f0 zeroP_dec0 eqCoef_dec ltC_dec multA0 invA0 a0 minusA0 divA0 
divA_dec0 ppc0 reduce1a unit = (fun x ->
 let res = 
  buch
    (spolyp f0 zeroP_dec0 ltC_dec multA0 invA0 a0 minusA0 divA0 ppc0)
    (divp_dec f0 zeroP_dec0 multA0 divA0 divA_dec0)
    (addEnd f0 zeroP_dec0) (addEnd f0 zeroP_dec0)
    (zerop_dec f0 zeroP_dec0) (ppcp f0 zeroP_dec0 ppc0)
    (nf f0 zeroP_dec0 ltC_dec multA0 invA0 a0 minusA0 divA0 reduce1a
      unit) (foreigner_dec f0 zeroP_dec0 eqCoef_dec multA0 ppc0) x
 in
  redd res Nil  (nf f0 zeroP_dec0 ltC_dec multA0 invA0 a0 minusA0 divA0 
reduce1a
      unit))
;;

let selectdiv f0 zeroP_dec0 multA0 divA0 divA_dec0 a0 q =
  list_rec Inright (fun a2 ->
    match a2 with
      Nil -> (fun l h' ->
        match h' with
          Inleft h'1 -> Inleft h'1
        | Inright -> Inright)
    | Cons (a3, l) -> (fun l0 h' ->
        match sumbool_rec Left_ren Right_ren (divA_dec0 a0 a3) with
          Left_ren -> Inleft (Cons (a3, l))
        | Right_ren ->
            (match h' with
               Inleft hyp1 -> Inleft hyp1
             | Inright -> Inright))) q
;;

let reduce1a f0 zeroP_dec0 ltC_dec plusA0 multA0 invA0 a0 minusA0 divA0 
divA_dec0 q p =
  well_founded_induction (fun x ->
    match x with
      Nil -> (fun h' -> Nil)
    | Cons (a2, l) -> (fun h' ->
        match selectdiv f0 zeroP_dec0 multA0 divA0 divA_dec0 a2 q with
          Inleft div1 ->
            (match div1 with
               Nil -> failwith "False_rec"
             | Cons (a3, l0) ->
                 h'
                   (spminusf f0 zeroP_dec0 ltC_dec multA0 invA0 a0
                     minusA0 divA0 a2 a3 l l0))
        | Inright -> Cons (a2, (h' l)))) p
;;

type 'd pol = ('d, mon) prod list;;

type 'd spol = 'd pol;;

let red eqD_dec b0 d0 plusD invD minusD multD ltC_dec b1 divD =
  reduce1a f (zeroP_dec b0 eqD_dec) ltC_dec (plusA plusD)
    (multA multD (mult_mon d0)) (invA invD) (Pair (b1, (zero_mon d0)))
    (minusA minusD) (divA divD (div_mon d0))
    (divA_dec b0 multD (mult_mon d0) divD (div_mon d0)
      (div_mon_clean d0))
;;

let unit eqD_dec b0 d0 b1 divD = function
  Nil -> Pair (b1, (zero_mon d0))
| Cons (a, l) ->
    (match a with
       Pair (co, c2) -> Pair ((divD b1 co), (zero_mon d0)))
;;

let eqA_nZ_dec b0 d0 multD divD a b =
  match a with
    Pair (x0, x) ->
      (match b with
         Pair (d1, m) ->
           (match div_mon_clean d0 x m with
              Pair (m1, b2) ->
                (match b2 with
                   True -> Left_ren
                 | False -> Right_ren)))
;;

let tbuch 
(* egalite des coef *) eqD_dec
(* coef nul *) b0 
(* dimension (Peano) *) d0
(* addition des coeff *) plusD
(* oppose d'un coef *) invD
(* difference des coeff *) minusD
(* multiplication de coef *) multD
(* ordre des monomes < > = *) ltC_dec 
(* coeff 1 *) b1
(* division des coef *) divD =
  buchf f (zeroP_dec b0 eqD_dec) (eqmon_dec d0) ltC_dec
    (multA multD (mult_mon d0)) (invA invD) (Pair (b1, (zero_mon d0)))
    (minusA minusD) (divA divD (div_mon d0))
    (eqA_nZ_dec b0 d0 multD divD) (ppc b1 (ppcm_mon d0))
    (red eqD_dec b0 d0 plusD invD minusD multD ltC_dec b1 divD)
    (unit eqD_dec b0 d0 b1 divD)
;;

let lplus eqD_dec b0 d0 plusD ltC_dec n =
  splus f (zeroP_dec b0 eqD_dec) ltC_dec (plusA plusD)
;;

let lmult eqD_dec b0 d0 plusD multD ltC_dec n =
  smult f (zeroP_dec b0 eqD_dec) ltC_dec (plusA plusD)
    (multA multD (mult_mon d0))
;;

let lscal eqD_dec b0 d0 multD h' h'0 =
  tmultm_lm (zeroP_dec b0 eqD_dec) (multA multD (mult_mon d0))
    (Pair (h', (zero_mon d0))) h'0
;;

let lpO eqD_dec b0 d0 =
  Nil
;;

let lp1 eqD_dec b0 d0 b1 =
  Cons ((Pair (b1, (zero_mon d0))), Nil)
;;

let lgen eqD_dec b0 d0 b1 n =
  Cons ((Pair (b1, (gen_mon d0 n))), Nil)
;;


