(*
    A parser for polynomials, with Maple-like syntax.

    Loic Pottier.
*)

open Genlex;;

let parse_lpol p0 p1 varP coefP plusP multP s =
   let rec powerP p n = match n with
       1 -> p
     | n -> (multP p (powerP p (n-1)))
   in
   let lexer = make_lexer ["^"; "+"; "-";"*";",";"[";"]"] in
   let rec parse_varexp = parser
      [< 'Ident v; e=parse_exp>] -> (powerP (varP v) e)
   and parse_exp = parser
      [< 'Kwd "^"; 'Int e >] -> e
    | [< >] -> 1
   and parse_mon = parser
      [<v = parse_varexp; m = parse_mon_rem >] -> (multP v m)
   and  parse_mon_rem  = parser
      [<'Kwd "*"; v = parse_varexp; m = parse_mon_rem >] -> (multP v m)
    | [<>] -> p1 
   and parse_pol= parser
    | [< 'Kwd "+"; p=parse_pol >] -> p
    | [< 'Int c; p = parse_pol1 c >] -> p
    | [< 'Kwd "-"; m = parse_mon ;p=parse_pol>] -> (plusP (multP (coefP (-1))
                                                                 m) p)
    | [< m = parse_mon ;p=parse_pol>] -> (plusP m p)
    | [<>] -> p0
   and parse_pol1 c = parser
    | [< 'Kwd "*"; m = parse_mon ;p=parse_pol >] 
        -> (plusP (multP (coefP c) m) p)
    | [< p=parse_pol>] -> if c=0 then p else (plusP (multP (coefP c) p1) p)
   and parse_lpol = parser
    | [< 'Kwd "["; p = parse_pol; l=parse_lpol_in; 'Kwd "]" >] -> p::l
   and parse_lpol_in = parser
    | [< 'Kwd ","; p = parse_pol; l=parse_lpol_in >] -> p::l
    | [< >] -> []
   in (parse_lpol (lexer(Stream.of_string s)))
;;

let parse_lvar s = 
   let lexer = make_lexer ["^"; "+"; "-";"*";",";"[";"]"] in
   let rec parse_lvar = parser
    | [< 'Kwd "["; 'Ident v; l=parse_lvar_in; 'Kwd "]" >] -> v::l
   and parse_lvar_in = parser
    | [< 'Kwd ","; 'Ident v; l=parse_lvar_in >] ->  v::l
    | [< >] -> []
   in (parse_lvar (lexer(Stream.of_string s)))
;;
